/*
 * CLCL
 *
 * fmt_unicode.c
 *
 * Copyright (C) 1996-2005 by Nakashima Tomoaki. All rights reserved.
 *		http://www.nakka.com/
 *		nakka@nakka.com
 */

/* Include Files */
#define _INC_OLE
#include <windows.h>
#undef  _INC_OLE

#include "CLCLPlugin.h"
#include "nedit.h"
#include "Memory.h"
#include "String.h"
#include "Message.h"

#include "resource.h"

/* Define */
#define TOOLTIP_SIZE				1024

#define	INI_FILE_NAME				TEXT("fmt_unicode.ini")

/* Global Variables */
HINSTANCE hInst;

TCHAR ini_path[BUF_SIZE];
int op_wordwrap;
TCHAR op_font_name[BUF_SIZE];
int op_font_size;
int op_font_weight;
int op_font_italic;
int op_font_charset;

static HANDLE hLib;
static HICON txt_icon;
static HWND hTxtWnd;

/* Local Function Prototypes */
static BOOL dll_initialize(void);

/*
 * DllMain - C
 */
int WINAPI DllMain(HINSTANCE hInstance, DWORD fdwReason, PVOID pvReserved)
{
	switch (fdwReason) {
	case DLL_PROCESS_ATTACH:
		hInst = hInstance;
		dll_initialize();
		break;

	case DLL_PROCESS_DETACH:
		break;

	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
		break;
	}
	return TRUE;
}

/*
 * dll_initialize - 
 */
static BOOL dll_initialize(void)
{
	TCHAR app_path[BUF_SIZE];
	TCHAR *p, *r;

	GetModuleFileName(hInst, app_path, BUF_SIZE - 1);
	for (p = r = app_path; *p != TEXT('\0'); p++) {
#ifndef UNICODE
		if (IsDBCSLeadByte((BYTE)*p) == TRUE) {
			p++;
			continue;
		}
#endif	// UNICODE
		if (*p == TEXT('\\') || *p == TEXT('/')) {
			r = p;
		}
	}
	*r = TEXT('\0');
	wsprintf(ini_path, TEXT("%s\\%s"), app_path, INI_FILE_NAME);

	op_wordwrap = GetPrivateProfileInt(TEXT("edit"), TEXT("wordwrap"), 1, ini_path);

	GetPrivateProfileString(TEXT("edit"), TEXT("font_name"), TEXT("Tahoma"), op_font_name, BUF_SIZE - 1, ini_path);
	op_font_size = GetPrivateProfileInt(TEXT("edit"), TEXT("font_size"), 9, ini_path);
	op_font_weight = GetPrivateProfileInt(TEXT("edit"), TEXT("font_weight"), 0, ini_path);
	op_font_italic = GetPrivateProfileInt(TEXT("edit"), TEXT("font_italic"), 0, ini_path);
	op_font_charset = GetPrivateProfileInt(TEXT("edit"), TEXT("font_charset"), 0, ini_path);
	return TRUE;
}

/*
 * file_read_buf - t@Cǂݍ
 */
BYTE *file_read_buf(const TCHAR *path, DWORD *ret_size, TCHAR *err_str)
{
	HANDLE hFile;
	DWORD size;
	DWORD ret;
	BYTE *buf;

	// t@CJ
	hFile = CreateFile(path, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == NULL || hFile == (HANDLE)-1) {
		message_get_error(GetLastError(), err_str);
		return NULL;
	}
	size = GetFileSize(hFile, NULL);
	if (size == 0xFFFFFFFF) {
		message_get_error(GetLastError(), err_str);
		CloseHandle(hFile);
		return NULL;
	}

	buf = (BYTE *)mem_alloc(size);
	if (buf == NULL) {
		message_get_error(GetLastError(), err_str);
		CloseHandle(hFile);
		return NULL;
	}

	// t@Cǂ݂
	if (ReadFile(hFile, buf, size, &ret, NULL) == FALSE) {
		message_get_error(GetLastError(), err_str);
		mem_free(&buf);
		CloseHandle(hFile);
		return NULL;
	}
	CloseHandle(hFile);

	if (ret_size != NULL) {
		*ret_size = size;
	}
	return buf;
}

/*
 * file_write_buf - t@Cɏ
 */
BOOL file_write_buf(const TCHAR *path, const BYTE *data, const DWORD size, TCHAR *err_str)
{
	HANDLE hFile;
	DWORD ret;

	// t@CJ
	hFile = CreateFile(path, GENERIC_READ | GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == NULL || hFile == (HANDLE)-1) {
		message_get_error(GetLastError(), err_str);
		return FALSE;
	}
	// t@C̏
	if (WriteFile(hFile, data, size, &ret, NULL) == FALSE) {
		message_get_error(GetLastError(), err_str);
		CloseHandle(hFile);
		return FALSE;
	}
	CloseHandle(hFile);
	return TRUE;
}

/*
 * get_format_header - `wb_̎擾
 */
__declspec(dllexport) BOOL CALLBACK get_format_header(const HWND hWnd, const int index, FORMAT_GET_INFO *fgi)
{
	switch (index) {
	case 0:
		strcpy(fgi->format_name, "UNICODE TEXT");
		strcpy(fgi->func_header, "unicode_");
		strcpy(fgi->comment, "UNICODE TEXT");
		return TRUE;
	}
	return FALSE;
}

/*
 * unicode_show_property - vpeB\
 */
__declspec(dllexport) BOOL CALLBACK unicode_show_property(const HWND hWnd)
{
	return FALSE;
}

/*
 * unicode_initialize - 
 */
__declspec(dllexport) BOOL CALLBACK unicode_initialize(void)
{
	if (txt_icon == NULL) {
		txt_icon = LoadImage(hInst, MAKEINTRESOURCE(IDI_ICON_UNICODE), IMAGE_ICON, 16, 16, 0);
	}
	nedit_regist(hInst);
	return TRUE;
}

/*
 * unicode_get_icon - `p̃ACR擾
 */
__declspec(dllexport) HICON CALLBACK unicode_get_icon(const int icon_size, BOOL *free_icon)
{
	*free_icon = FALSE;
	return txt_icon;
}

/*
 * unicode_free - I
 */
__declspec(dllexport) BOOL CALLBACK unicode_free(void)
{
	if (txt_icon != NULL) {
		DestroyIcon(txt_icon);
		txt_icon = NULL;
	}
	return TRUE;
}

/*
 * unicode_initialize_item - ACȅ
 */
__declspec(dllexport) BOOL CALLBACK unicode_initialize_item(DATA_INFO *di, const BOOL set_init_data)
{
	return FALSE;
}

/*
 * unicode_copy_data - f[^̃Rs[
 */
__declspec(dllexport) HANDLE CALLBACK unicode_copy_data(const TCHAR *format_name, const HANDLE data, DWORD *ret_size)
{
	return NULL;
}

/*
 * unicode_data_to_bytes - f[^oCgɕϊ
 */
__declspec(dllexport) BYTE* CALLBACK unicode_data_to_bytes(const DATA_INFO *di, DWORD *ret_size)
{
	return NULL;
}

/*
 * unicode_bytes_to_data - oCgf[^ɕϊ
 */
__declspec(dllexport) HANDLE CALLBACK unicode_bytes_to_data(const TCHAR *format_name, const BYTE *data, DWORD *size)
{
	return NULL;
}

/*
 * unicode_get_file_info - R_CAO̎擾
 */
__declspec(dllexport) int CALLBACK unicode_get_file_info(const TCHAR *format_name, const DATA_INFO *di, OPENFILENAME *of, const BOOL mode)
{
	of->lpstrFilter = (TCHAR *)"*.txt\0*.txt\0*.*\0*.*\0\0";
	of->nFilterIndex = 1;
	of->lpstrDefExt = (TCHAR *)"txt";
	return 1;
}

/*
 * unicode_data_to_file - f[^t@Cɕۑ
 */
__declspec(dllexport) BOOL CALLBACK unicode_data_to_file(DATA_INFO *di, const char *file_name, const int filter_index, char *err_str)
{
	WCHAR path[BUF_SIZE];
	WCHAR err[BUF_SIZE];
	BYTE *tmp;

	if (di->data == NULL) {
		return FALSE;
	}
	if ((tmp = GlobalLock(di->data)) == NULL) {
		message_get_error(GetLastError(), err);
		WideCharToMultiByte(CP_ACP, 0, err, -1, err_str, BUF_SIZE - 1, NULL, NULL);
		return FALSE;
	}
	// t@Cɏ
	char_to_tchar(file_name, path, BUF_SIZE - 1);
	if (file_write_buf(path, tmp, di->size - sizeof(WCHAR), err) == FALSE) {
		GlobalUnlock(di->data);
		WideCharToMultiByte(CP_ACP, 0, err, -1, err_str, BUF_SIZE - 1, NULL, NULL);
		return FALSE;
	}
	GlobalUnlock(di->data);
	return TRUE;
}

/*
 * unicode_file_to_data - t@Cf[^쐬
 */
__declspec(dllexport) HANDLE CALLBACK unicode_file_to_data(const char *file_name, const char *format_name, DWORD *ret_size, char *err_str)
{
	HANDLE ret = NULL;
	WCHAR *data;
	WCHAR *mem;
	WCHAR path[BUF_SIZE];
	WCHAR err[BUF_SIZE];
	DWORD size;

	// t@C̓ǂݍ
	char_to_tchar(file_name, path, BUF_SIZE - 1);
	data = (WCHAR *)file_read_buf(path, &size, err);
	if (data == NULL) {
		WideCharToMultiByte(CP_ACP, 0, err, -1, err_str, BUF_SIZE - 1, NULL, NULL);
		return NULL;
	}
	// Rs[m
	if ((ret = GlobalAlloc(GHND, size + sizeof(WCHAR))) == NULL) {
		message_get_error(GetLastError(), err);
		mem_free(&data);
		WideCharToMultiByte(CP_ACP, 0, err, -1, err_str, BUF_SIZE - 1, NULL, NULL);
		return NULL;
	}
	// Rs[惍bN
	if ((mem = GlobalLock(ret)) == NULL) {
		message_get_error(GetLastError(), err);
		GlobalFree(ret);
		mem_free(&data);
		WideCharToMultiByte(CP_ACP, 0, err, -1, err_str, BUF_SIZE - 1, NULL, NULL);
		return NULL;
	}
	// Rs[
	CopyMemory(mem, data, size);
	*(mem + (size / sizeof(WCHAR))) = L'\0';
	if (ret_size != NULL) {
		*ret_size = size + sizeof(WCHAR);
	}
	// bN
	GlobalUnlock(ret);
	mem_free(&data);
	return ret;
}

/*
 * unicode_free_data - f[^̉
 */
__declspec(dllexport) BOOL CALLBACK unicode_free_data(const TCHAR *format_name, HANDLE data)
{
	if (data == NULL) {
		return TRUE;
	}
	if (GlobalFree((HGLOBAL)data) != NULL) {
		return FALSE;
	}
	return TRUE;
}

/*
 * unicode_free_item - ACẻ
 */
__declspec(dllexport) BOOL CALLBACK unicode_free_item(DATA_INFO *di)
{
	return FALSE;
}

/*
 * unicode_get_menu_title - j[^Cg̎擾
 */
__declspec(dllexport) BOOL CALLBACK unicode_get_menu_title(DATA_INFO *di)
{
	WCHAR *mem;
	char buf[BUF_SIZE];
	char tmp[BUF_SIZE];
	char *p, *r;
	int size;

	if (di->data == NULL || (mem = GlobalLock(di->data)) == NULL) {
		return FALSE;
	}
	if (*mem == L'\0') {
		GlobalUnlock(di->data);
		return TRUE;
	}

	// j[p
	WideCharToMultiByte(CP_ACP, 0, mem, di->size / sizeof(WCHAR), tmp, BUF_SIZE - 1, NULL, NULL);
	GlobalUnlock(di->data);

	p = tmp;
	size = di->size / sizeof(WCHAR);
	r = buf;
	for (; size > 0 && (*p == ' ' || *p == '\t' || *p == '\r' || *p == '\n'); p++)
		;
	while ((r - buf) < (BUF_SIZE - 4) && size > 0 && *p != '\0') {
		if (*p == '\r' || *p == '\n') {
			strcpy(r, "...");
			r += 3;
			break;
		}
		switch (*p) {
		case ' ':
		case '\t':
			(*r++) = ' ';
			for (; size > 0 && (*p == ' ' || *p == '\t'); p++, size--)
				;
			break;

		case '&':
			(*r++) = '&';
			(*r++) = *(p++);
			size--;
			break;

		default:
			(*r++) = *(p++);
			size--;
			break;
		}
	}
	*r = '\0';

	if (*buf != '\0') {
		di->menu_title = mem_alloc(sizeof(char) * ((r - buf) + 1));
		if (di->menu_title == NULL) {
			return FALSE;
		}
		CopyMemory(di->menu_title, buf, sizeof(char) * ((r - buf) + 1));
		di->free_title = TRUE;
	}
	return TRUE;
}

/*
 * unicode_get_menu_icon - j[pACR̎擾
 */
__declspec(dllexport) BOOL CALLBACK unicode_get_menu_icon(DATA_INFO *di, const int icon_size)
{
	di->menu_icon = txt_icon;
	di->free_icon = FALSE;
	return TRUE;
}

/*
 * unicode_get_menu_bitmap - j[prbg}bv̎擾
 */
__declspec(dllexport) BOOL CALLBACK unicode_get_menu_bitmap(DATA_INFO *di, const int width, const int height)
{
	return FALSE;
}

/*
 * unicode_get_tooltip_text - j[pc[`bveLXg
 */
__declspec(dllexport) char* CALLBACK unicode_get_tooltip_text(DATA_INFO *di)
{
	char *ret;
	WCHAR *mem;
	int len;

	if (di->data == NULL || (mem = GlobalLock(di->data)) == NULL) {
		return NULL;
	}
	if (*mem == L'\0') {
		GlobalUnlock(di->data);
		return NULL;
	}
	ret = mem_alloc(sizeof(char) * (TOOLTIP_SIZE + 1));
	if (ret == NULL) {
		GlobalUnlock(di->data);
		return NULL;
	}
	len = (di->size / sizeof(WCHAR) < (DWORD)TOOLTIP_SIZE) ?
		di->size / sizeof(WCHAR) : TOOLTIP_SIZE;
	WideCharToMultiByte(CP_ACP, 0, (WCHAR *)mem, di->size / sizeof(WCHAR), ret, TOOLTIP_SIZE, NULL, NULL);
	*(ret + len) = '\0';
	GlobalUnlock(di->data);
	return ret;
}

/*
 * unicode_window_create - f[^\EBhE̍쐬
 */
__declspec(dllexport) HWND CALLBACK unicode_window_create(const HWND parent_wnd)
{
	if (hTxtWnd == NULL) {
		hTxtWnd = CreateWindowEx(WS_EX_CLIENTEDGE, NEDIT_WND_CLASS, NULL,
			WS_TABSTOP | WS_CHILD | WS_VISIBLE | WS_VSCROLL | ES_MULTILINE | ES_NOHIDESEL |
			((op_wordwrap == 0) ? WS_HSCROLL : 0),
			0, 0, 0, 0, parent_wnd, (HMENU)0, hInst, NULL);
	}
	return hTxtWnd;
}

/*
 * unicode_window_destroy - f[^\EBhE̔j
 */
__declspec(dllexport) BOOL CALLBACK unicode_window_destroy(const HWND hWnd)
{
	hTxtWnd = NULL;
	return TRUE;
}

/*
 * unicode_window_show_data - f[^̕\
 */
__declspec(dllexport) BOOL CALLBACK unicode_window_show_data(const HWND hWnd, DATA_INFO *di, const BOOL lock)
{
	WCHAR *mem;

	if (di->data == NULL) {
		SendMessage(hWnd, WM_SETTEXT, 0, (LPARAM)TEXT(""));
	} else {
		if ((mem = GlobalLock(di->data)) == NULL) {
			return FALSE;
		}
		SendMessage(hWnd, WM_SETTEXT, di->size / sizeof(WCHAR), (LPARAM)mem);
		GlobalUnlock(di->data);
	}
	SendMessage(hWnd, EM_SETREADONLY, lock, 0);
	return TRUE;
}

/*
 * unicode_window_save_data - f[^̕ۑ
 */
__declspec(dllexport) BOOL CALLBACK unicode_window_save_data(const HWND hWnd, DATA_INFO *di)
{
	HANDLE data;
	WCHAR *to_mem;
	DWORD size;

	if (hWnd == NULL ||
		(SendMessage(hWnd, EM_GETREADONLY, 0, 0) == TRUE) ||
		SendMessage(hWnd, EM_GETMODIFY, 0, 0) == FALSE) {
		return FALSE;
	}
	if (di->data != NULL) {
		GlobalFree(di->data);
		di->data = NULL;
		di->size = 0;
	}
	size = SendMessage(hWnd, WM_GETTEXTLENGTH, 0, 0);
	// f[^̍쐬
	if ((data = GlobalAlloc(GHND, sizeof(WCHAR) * (size + 1))) == NULL) {
		return FALSE;
	}
	if ((to_mem = GlobalLock(data)) == NULL) {
		GlobalFree(data);
		return FALSE;
	}
	SendMessage(hWnd, WM_GETTEXT, size + 1, (LPARAM)to_mem);
	GlobalUnlock(data);

	// Vf[^ݒ
	di->data = data;
	di->size = sizeof(WCHAR) * (size + 1);

	SendMessage(hWnd, EM_SETMODIFY, FALSE, 0);
	return TRUE;
}

/*
 * unicode_window_hide_data - f[^̔\
 */
__declspec(dllexport) BOOL CALLBACK unicode_window_hide_data(const HWND hWnd, DATA_INFO *di)
{
	SendMessage(hWnd, WM_SETTEXT, 0, (LPARAM)TEXT(""));
	return TRUE;
}
/* End of source */
